/*
 * Created by IntelliJ IDEA.
 * User: Dominik
 * Date: 10.05.2004
 * Time: 16:32:16
 */

import java.awt.*;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * A light ray, whose spectrum can consist of several wavelengths.
 */
public class Ray {
    /**
     * Constructs an empty ray in the vacuum.
     * @param pos  the start position of the ray
     * @param dir  the direction of the ray (length 1 corresponds with the speed of light in vacuum)
     */
    public Ray(Vector2D pos, Vector2D dir) {
        this.pos = new Vector2D(pos);
        this.dir = new Vector2D(dir);
        history = new ArrayList();
        addToHistory(pos);
        wavelengths = new ArrayList();
        active = true;
        vacuum = true;
    }

    /**
     * Constructs a ray in the vaccum, whose spectrum consists of one single, given wavelength.
     * @param pos  the start position of the ray
     * @param dir  the direction of the ray (length 1 corresponds with the speed of light in vacuum)
     * @param wl  the wavelength of the ray
     */
    public Ray(Vector2D pos, Vector2D dir, double wl) {
        this(pos, dir);
        addWavelength(wl);
    }

    /**
     * Returns the current position of the ray.
     */
    public Vector2D getPos() {
        return pos;
    }

    /**
     * Sets a new position.
     */
    public void setPos(Vector2D v) {
        pos = new Vector2D( v );
    }

    /**
     * Returns the current direction of the ray.
     */
    public Vector2D getDir() {
        return dir;
    }

    /**
     * Returns true iff the ray is active.
     */
    public boolean isActive() {
        return active;
    }

    /**
     * Returns true iff the ray consists of exactly one wavelength.
     */
    public boolean isMonochromatic() {
        return (wavelengths.size() == 1);
    }

    /**
     * Deactivates the ray. That means, it will stop moving.
     */
    public void deactivate() {
        active = false;
    }

    /**
     * Add a wavelength to the spectrum of this ray.
     */
    public void addWavelength(double wl) {
        wavelengths.add(new Double(wl));
    }

    /**
     * Returns the first wavelength of this ray.
     * This method is useful when dealing with monochromatic rays.
     */
    public double getWavelength() {
        return ((Double)wavelengths.get(0)).doubleValue();
    }

    /**
     * Returns an iterator for all wavelengths of this ray.
     */
    public Iterator getWavelengths() {
        return wavelengths.iterator();
    }

    /**
     * Adds a point to the history of this ray.
     * The history stores all points at which the ray changes its direction.
     */
    public void addToHistory(Vector2D v) {
        history.add( new Vector2D(v) );
    }

    /**
     * Draws this ray on the given canvas.
     */
    public void draw(MyCanvas c, Graphics g) {
        // determine the color (sum of the colors of all wavelengths)
        MyColor color = new MyColor();
        Iterator it = getWavelengths();
        while (it.hasNext()) {
            double wl = ((Double)it.next()).doubleValue();
            color.add( new MyColor((int)wl) );
        }
        g.setColor( color.getColor() );

        // follow the history of the ray and draw it
        Vector2D start = pos;
        for (int i=history.size()-1; i>=0; i--) {
            Vector2D end = (Vector2D) history.get(i);
            g.drawLine( c.pX(start), c.pY(start), c.pX(end), c.pY(end) );
            start = end;
        }
    }

    /**
     * Returns the position at which this ray would be after moving for the given time interval.
     */
    public Vector2D getNextPoint(double time) {
        return pos.getSum( dir.getProduct(time) );
    }

    /**
     * Moves this ray for the given time interval without changing its direction.
     */
    public void moveStraight(double time) {
        pos = getNextPoint(time);
    }

    /**
     * Refracts this ray.
     * @param refractionIndex  the index of refraction
     * @param incoming  vector that points from the point of intersection to the last position of this ray
     * @param edge  vector that represents the edge of the prism that the ray intersects
     */
    public void refract(double refractionIndex, Vector2D incoming, Vector2D edge) {
        if ( ! vacuum ) refractionIndex = 1 / refractionIndex;
        vacuum = !vacuum;

        incoming.normalize();
        edge.normalize();

        double alpha = Math.PI/2 - Math.acos( incoming.getScalarProduct(edge) );
        double beta = Math.asin( Math.sin(alpha) / refractionIndex );
        double rotation = Math.PI/2 + beta;
        if (Vector2D.getOrientation(edge, new Vector2D(), incoming) > 0) rotation *= (-1);

        /*
        System.out.println(refractionIndex);
        System.out.println(alpha);
        System.out.println( Math.sin(alpha) / refractionIndex );
        System.out.println(beta);
        System.out.println();
        */

        dir.set(edge).rotate(rotation);

        dir.multiply(1/refractionIndex);
    }

    private Vector2D pos;
    private Vector2D dir;
    private ArrayList history;
    private ArrayList wavelengths;
    private boolean active;
    private boolean vacuum;
}
