/*
 * Created by IntelliJ IDEA.
 * User: Dominik
 * Date: 10.05.2004
 * Time: 19:53:41
 */

import java.util.ArrayList;
import java.util.Iterator;
import java.awt.*;

/**
 * The scene consisting of prisms and detectors.
 */
public class Scene {
    /**
     * The constructor.
     */
    public Scene(Output output) {
        prisms = new ArrayList();
        detectors = new ArrayList();
        this.output = output;
    }

    /**
     * Returns the output object.
     */
    public Output getOutput() { return output; }

    /**
     * Adds the given prism to the scene.
     */
    public void addPrism(Prism prism) {
        prisms.add( prism );
    }

    /**
     * Adds the given detector to the scene.
     */
    public void addDetector(Detector detector) {
        detectors.add( detector );
    }

    /**
     * Resets an ArrayList that is used for the temporary storage of rays that arise due to dispersion.
     */
    public void resetNewRays() {
        newRays = new ArrayList();
    }

    /**
     * Returns the ArrayList that is used for the temporary storage of rays that arise due to dispersion.
     */
    public ArrayList getNewRays() {
        return newRays;
    }

    /**
     * Returns true iff the given ray, moved for the given time interval, would intersect with a prism.
     * If there is an intersection, refraction and dispersion is taken into consideration.
     */
    boolean isIntersectingPrism(Ray ray, double time) {
        for (int i=0; i<prisms.size(); i++) {
            Prism prism = (Prism)prisms.get(i);
            Vector2D edge = new Vector2D();
            // determine the point of intersection with the current prism
            Vector2D intersection = prism.getIntersection( ray.getPos(), ray.getNextPoint(time), edge );
            if (intersection != null) {
                Vector2D incoming = new Vector2D(ray.getPos());
                incoming.subtract( intersection );
                // compute the time interval until the ray hits the prism
                double partTime = 0;
                if (ray.getDir().x() != 0) {
                    partTime = (intersection.x() - ray.getPos().x()) / ray.getDir().x();
                }
                else {
                    partTime = (intersection.y() - ray.getPos().y()) / ray.getDir().y();
                }
                if ( ray.isMonochromatic() ) {
                    // case 1: monochromatic ==> just refract the ray
                    ray.addToHistory( intersection );
                    ray.setPos( intersection );
                    ray.refract( prism.getRefractionIndex( ray.getWavelength() ), incoming, edge );
                    ray.moveStraight( time - partTime );
                }
                else {
                    // case 2: more than one wavelength ==> split the ray into several monochromatic rays and
                    // refract each of them independently.
                    ray.setPos( intersection );
                    ray.deactivate(); // deactivate the original ray
                    Iterator it = ray.getWavelengths();
                    while (it.hasNext()) {
                        double wl = ((Double) it.next()).doubleValue();
                        Ray newRay = new Ray( intersection, ray.getDir(), wl);
                        newRay.refract( prism.getRefractionIndex(wl), incoming, edge );
                        newRay.moveStraight( time - partTime );
                        newRays.add( newRay );
                    }
                }
                return true;
            }
        }
        return false;
    }

    /**
     * Returns true iff the given ray, moved for the given time interval, would intersect with a detector.
     * If there is an intersection, the ray is added to the output and deactivated.
     */
    boolean isIntersectingDetector(Ray ray, double time) {
        for (int i=0; i<detectors.size(); i++) {
            Detector detector = (Detector)detectors.get(i);
            Vector2D intersection = detector.getIntersection( ray.getPos(), ray.getNextPoint(time) );
            if (intersection != null) {
                ray.setPos( intersection );
                ray.deactivate();
                output.addWavelength( ray.getWavelength() );
                return true;
            }
        }
        return false;
    }

    /**
     * Draws the scene.
     */
    public void draw(MyCanvas c, Graphics g) {
        g.setColor(Color.DARK_GRAY);

        // draw the prisms
        for (int i=0; i<prisms.size(); i++) {
            Prism p = (Prism)prisms.get(i);
            int[] x = { c.pX(p.a()), c.pX(p.b()), c.pX(p.c())};
            int[] y = { c.pY(p.a()), c.pY(p.b()), c.pY(p.c())};
            g.fillPolygon(x,y,3);
        }

        // draw the detectors
        for (int i=0; i<detectors.size(); i++) {
            Detector d = (Detector)detectors.get(i);
            double thickness = 10;
            Vector2D normal = d.b().getDifference(d.a()).getNormal().setLength(thickness);

            int[] x = { c.pX(d.a().getDifference(normal)), c.pX(d.b().getDifference(normal)),
                        c.pX(d.b().getSum(normal)), c.pX(d.a().getSum(normal))};
            int[] y = { c.pY(d.a().getDifference(normal)), c.pY(d.b().getDifference(normal)),
                        c.pY(d.b().getSum(normal)), c.pY(d.a().getSum(normal))};
            g.fillPolygon(x,y,4);
        }
    }

    private ArrayList prisms;
    private ArrayList newRays;
    private ArrayList detectors;
    private Output output;
}
