/*
 * Created by IntelliJ IDEA.
 * User: Dominik
 * Date: 10.05.2004
 * Time: 16:36:32
 */

/**
 * A 2-dimensional double vector.
 */
public class Vector2D {
    /**
     * Constructs a 0-vector.
     */
    public Vector2D() {
        this(0,0);
    }

    /**
     * Construcs a vector with the given coordinates.
     */
    public Vector2D(double x, double y) {
        this.x = x;
        this.y = y;
    }

    /**
     * Constructs a vector with the given coordinates.
     * If requested, the vector is normalized so that its length = 1.
     */
    public Vector2D(double x, double y, boolean normalize) {
        this(x,y);
        if (normalize) normalize();
    }

    /**
     * Constructs a vector that is equal to a given vector.
     */
    public Vector2D(Vector2D v) {
        set(v);
    }

    /**
     * Returns a String representation of this vector.
     */
    public String toString() {
        return x + " / " + y;
    }

    /**
     * Sets the coordinates of this vector to the coordinates of a given vector and
     * returns this vector.
     */
    public Vector2D set(Vector2D v) {
        x = v.x();
        y = v.y();
        return this;
    }

    /**
     * Returns the x-coordinate of this vector.
     */
    public double x() {return x;}

    /**
     * Returns the y-coordinate of this vector.
     */
    public double y() {return y;}

    /**
     * Normalizes this vector and returns this vector.
     */
    public Vector2D normalize() {
        return setLength(1);
    }

    /**
     * Sets the length of this vector to the given value and
     * returns this vector.
     */
    public Vector2D setLength(double length) {
        double oldLength = getLength();
        x *= length / oldLength;
        y *= length / oldLength;
        return this;
    }

    /**
     * Returns the length of this vector.
     */
    public double getLength() {
        return Math.sqrt(x*x+y*y);
    }

    /**
     * Returns the scalar product of this vector with a given vector.
     */
    public double getScalarProduct(Vector2D v) {
        return x*v.x() + y*v.y();
    }

    /**
     * Returns the product of this vector with a given scalar.
     */
    public Vector2D getProduct(double s) {
        return new Vector2D(x*s, y*s);
    }

    /**
     * Multiplies this vector with a given scalar.
     */
    public void multiply(double s) {
        x *= s;
        y *= s;
    }

    /**
     * Returns the sum of this vector and a given vector.
     */
    public Vector2D getSum(Vector2D v) {
        return new Vector2D( x + v.x(), y + v.y() );
    }

    /**
     * Adds a given vector to this vector.
     */
    public void add(Vector2D v) {
        x += v.x();
        y += v.y();
    }

    /**
     * Returns the difference between this vector and a given vector.
     */
    public Vector2D getDifference(Vector2D v) {
        return new Vector2D( x - v.x(), y - v.y() );
    }

    /**
     * Subtracts a given vector from this vector.
     */
    public Vector2D subtract(Vector2D v) {
        x -= v.x();
        y -= v.y();
        return this;
    }

    /**
     * Returns the normal of this vector.
     * I.e., a vector that is perpendicular to this vector.
     */
    public Vector2D getNormal() {
        return new Vector2D(y, -x);
    }

    /**
     * Rotates this vector by a given angle.
     */
    public Vector2D rotate(double angle) {
        double cos = Math.cos(-angle);
        double sin = Math.sin(-angle);

        double newX = x * cos - y * sin;
        y           = x * sin + y * cos;
        x = newX;
        return this;
    }


    private double x;
    private double y;



    /**
     * Returns a value > 0 if p, q, r form a left turn.
     * Returns 0 if p, q, r are collinear.
     * Returns a value < 0 if p, q, r form a right turn.
     */
    public static double getOrientation(Vector2D p, Vector2D q, Vector2D r) {
        return q.x*r.y + p.x*q.y + p.y*r.x - q.x*p.y - r.x*q.y - r.y*p.x;
    }

    /**
     * Returns the point of intersection between two given lines or null if there is no intersection.
     */
    public static Vector2D getIntersectionOfTwoLines(Vector2D a, Vector2D b, Vector2D c, Vector2D d) {
        if (! ( (Vector2D.getOrientation(a,b,c) * Vector2D.getOrientation(a,b,d) <= 0) &&
                (Vector2D.getOrientation(c,d,a) * Vector2D.getOrientation(c,d,b) <= 0) ))
            return null;

        double r = d.y() - b.y() -  (d.x()-b.x())*(a.y()-b.y())/(a.x()-b.x());
        r /=       d.y() - c.y() -  (d.x()-c.x())*(a.y()-b.y())/(a.x()-b.x());

        Vector2D result = new Vector2D( c.getProduct(r) );
        result.add( d.getProduct(1-r) );
        return result;
    }

}
